import httmock
import os
import sys
import threading
import unittest
from . import captureOutput
from girder_worker.core.io import (make_stream_push_adapter,
                                   make_stream_fetch_adapter)
from girder_worker.core.utils import run_process
from six.moves import BaseHTTPServer, socketserver

_iscript = os.path.join(os.path.dirname(__file__), 'stream_input.py')
_oscript = os.path.join(os.path.dirname(__file__), 'stream_output.py')
_pipepath = os.path.join(os.path.dirname(__file__), 'namedpipe')
_socket_port = int(os.environ.get('WORKER_TEST_SOCKET_PORT', 7941))
_server = None
_req_chunks = []


class Handler(BaseHTTPServer.BaseHTTPRequestHandler):
    def do_PUT(self):
        while True:
            length = int(self.rfile.readline(), 16)
            if length:
                _req_chunks.append((length, self.rfile.read(length)))
            else:
                break
            self.rfile.readline()  # read the empty line between chunks

        self.close_connection = True
        self.send_response(200, message='received')
        self.end_headers()

    def log_message(self, *args, **kwargs):
        pass  # Override so we don't print to stderr


class ServerThread(object):
    def __init__(self, port):
        self.port = port
        self.httpd = socketserver.TCPServer(('127.0.0.1', self.port), Handler)
        self.thread = threading.Thread(target=self.start, args=(self.httpd,))
        self.thread.start()

    def start(self, httpd):
        httpd.serve_forever()

    def stop(self):
        self.httpd.shutdown()
        self.httpd.server_close()
        self.thread.join(timeout=10)


def setUpModule():
    global _server
    _server = ServerThread(_socket_port)


def tearDownModule():
    if _server:
        _server.stop()


class TestStream(unittest.TestCase):
    def setUp(self):
        super(TestStream, self).setUp()
        if os.path.exists(_pipepath):
            os.unlink(_pipepath)
        os.mkfifo(_pipepath)

    def tearDown(self):
        super(TestStream, self).tearDown()
        if os.path.exists(_pipepath):
            os.unlink(_pipepath)

    def testOutputStreams(self):
        output_spec = {
            'mode': 'http',
            'method': 'PUT',
            'url': 'http://localhost:%d' % _socket_port
        }

        fd = os.open(_pipepath, os.O_RDONLY | os.O_NONBLOCK)
        adapters = {
            fd: make_stream_push_adapter(output_spec)
        }
        cmd = [sys.executable, _oscript, _pipepath]

        try:
            with captureOutput() as stdpipes:
                run_process(cmd, adapters)
        except Exception:
            print('Stdout/stderr from exception: ')
            print(stdpipes)
            raise
        self.assertEqual(stdpipes, ['start\ndone\n', ''])
        self.assertEqual(len(_req_chunks), 1)
        self.assertEqual(_req_chunks[0], (9, 'a message'))

    def testInputStreams(self):
        input_spec = {
            'mode': 'http',
            'method': 'GET',
            'url': 'http://mockedhost'
        }

        @httmock.urlmatch(netloc='^mockedhost$', method='GET')
        def mock_fetch(url, request):
            return 'hello\nworld'

        adapters = {
            _pipepath: make_stream_fetch_adapter(input_spec)
        }
        cmd = [sys.executable, _iscript, _pipepath]

        try:
            with captureOutput() as stdpipes, httmock.HTTMock(mock_fetch):
                run_process(cmd, input_pipes=adapters)
        except Exception:
            print('Stdout/stderr from exception: ')
            print(stdpipes)
            raise
        self.assertEqual(stdpipes, ['olleh\ndlrow\n', ''])



from __future__ import print_function
from ConfigParser import ConfigParser

import argparse
import os
import pip
import sys


def main(args):
    basePath = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

    # Read the configuration files
    files = ('worker.dist.cfg', 'worker.local.cfg')
    config = ConfigParser()
    config.read([os.path.join(basePath, 'girder_worker', f) for f in files])
    os.chdir(basePath)
    plugins = os.environ.get(
        'GIRDER_WORKER_PLUGINS_ENABLED',
        config.get('girder_worker', 'plugins_enabled'))
    plugins = [p.strip() for p in plugins.split(',') if p.strip()]

    installArgs = ['requirements.in']
    if args.mode in ('dev', 'devel', 'development'):
        installArgs.append('requirements-dev.in')

    # Install plugins requirements files
    pluginsDir = os.path.join('girder_worker', 'plugins')
    for path in os.listdir(pluginsDir):
        if args.all or path in plugins:
            pluginReqs = os.path.join(pluginsDir, path, 'requirements.txt')
            if os.path.exists(os.path.join(basePath, pluginReqs)):
                installArgs.append(pluginReqs)

    pipArgs = ['install'] + [
        os.path.join('-r' + basePath, req) for req in installArgs
    ]

    print('\033[32m*** Installing: %s\033[0m' % ' '.join(installArgs))
    if pip.main(pipArgs) != 0:
        print('\033[1;91m*** Error installing %s, stopping.\033[0m' % file,
              file=sys.stderr)
        sys.exit(1)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='Install required pip packages.')
    parser.add_argument('-m', '--mode', default='prod',
                        help='install for "dev" or "prod" (the default)')
    parser.add_argument('-a', '--all', help='install requirements of all '
                        'plugins rather than just enabled ones',
                        action='store_true')
    args = parser.parse_args()

    main(args)



from girder_worker.tasks import run
import unittest


class TestNumber(unittest.TestCase):

    def setUp(self):
        self.analysis = {
            'name': 'add',
            'inputs': [
                {
                    'name': 'a',
                    'type': 'number',
                    'format': 'number',
                    'default': {
                        'format': 'json',
                        'data': '0'
                    }
                },
                {
                    'name': 'b',
                    'type': 'number',
                    'format': 'number'
                }
            ],
            'outputs': [{'name': 'c', 'type': 'number', 'format': 'number'}],
            'script': 'c = a + b',
            'mode': 'python'
        }

    def test_numeric(self):
        outputs = run(
            self.analysis,
            inputs={
                'a': {'format': 'number', 'data': 1},
                'b': {'format': 'number', 'data': 2}
            },
            outputs={
                'c': {'format': 'number'}
            })
        self.assertEqual(outputs['c']['format'], 'number')
        self.assertEqual(outputs['c']['data'], 3)

    def test_json(self):
        outputs = run(
            self.analysis,
            inputs={
                'a': {'format': 'json', 'data': '1'},
                'b': {'format': 'json', 'data': '2'}
            },
            outputs={
                'c': {'format': 'json'}
            })
        self.assertEqual(outputs['c']['format'], 'json')
        self.assertEqual(outputs['c']['data'], '3')

    def test_float(self):
        outputs = run(
            self.analysis,
            inputs={
                'a': {'format': 'number', 'data': 1.5},
                'b': {'format': 'number', 'data': 2.5}
            },
            outputs={
                'c': {'format': 'number'}
            })
        self.assertEqual(outputs['c']['format'], 'number')
        self.assertEqual(outputs['c']['data'], 4)

    def test_default(self):
        outputs = run(
            self.analysis,
            inputs={
                'b': {'format': 'number', 'data': 2}
            },
            outputs={
                'c': {'format': 'number'}
            })
        self.assertEqual(outputs['c']['format'], 'number')
        self.assertEqual(outputs['c']['data'], 2)

        self.assertRaisesRegexp(
            Exception, "^Required input 'b' not provided.$",
            run, self.analysis,
            inputs={
                'a': {'format': 'number', 'data': 2}
            },
            outputs={
                'c': {'format': 'number'}
            })


if __name__ == '__main__':
    unittest.main()



import celery
from celery.result import AsyncResult

from girder_worker.context import get_context
from girder_worker.utils import is_builtin_celery_task, is_revoked

from girder_worker_utils import _walk_obj
from girder_worker_utils.decorators import describe_function
import six


class GirderAsyncResult(AsyncResult):
    def __init__(self, *args, **kwargs):
        self._job = None
        super(GirderAsyncResult, self).__init__(*args, **kwargs)

    @property
    def job(self):
        context = get_context()
        if self._job is None:
            self._job = context.get_async_result_job_property(self)
        return self._job


class Task(celery.Task):
    """Girder Worker Task object"""

    _girder_job_title = '<unnamed job>'
    _girder_job_type = 'celery'
    _girder_job_public = False
    _girder_job_handler = 'celery_handler'
    _girder_job_other_fields = {}

    @classmethod
    def girder_job_defaults(cls):
        return {
            'girder_job_title': cls._girder_job_title,
            'girder_job_type': cls._girder_job_type,
            'girder_job_public': cls._girder_job_public,
            'girder_job_handler': cls._girder_job_handler,
            'girder_job_other_fields': cls._girder_job_other_fields
        }

    # These keys will be removed from apply_async's kwargs or options and
    # transfered into the headers of the message.
    reserved_headers = [
        'girder_client_token',
        'girder_api_url',
        'girder_result_hooks']

    # These keys will be available in the 'properties' dictionary inside
    # girder_before_task_publish() but will not be passed along in the message
    reserved_options = [
        'girder_user',
        'girder_job_title',
        'girder_job_type',
        'girder_job_public',
        'girder_job_handler',
        'girder_job_other_fields']

    def AsyncResult(self, task_id, **kwargs):
        return GirderAsyncResult(task_id, backend=self.backend,
                                 task_name=self.name, app=self.app, **kwargs)

    def apply_async(self, args=None, kwargs=None, task_id=None, producer=None,
                    link=None, link_error=None, shadow=None, **options):

        if is_builtin_celery_task(self.name):
            return super(Task, self).apply_async(
                args=args, kwargs=kwargs, task_id=task_id, producer=producer,
                link=link, link_error=link_error, shadow=shadow, **options)

        # Pass girder related job information through to
        # the signals by adding this information to options['headers']
        # This sets defaults for reserved_options based on the class defaults,
        # or values defined by the girder_job() dectorator
        headers = {
            'girder_job_title': self._girder_job_title,
            'girder_job_type': self._girder_job_type,
            'girder_job_public': self._girder_job_public,
            'girder_job_handler': self._girder_job_handler,
            'girder_job_other_fields': self._girder_job_other_fields,
        }

        # Certain keys may show up in either kwargs (e.g. via
        # .delay(girder_token='foo') or in options (e.g.
        # .apply_async(args=(), kwargs={}, girder_token='foo') For
        # those special headers, pop them out of kwargs or options and
        # put them in headers so they can be picked up by the
        # before_task_publish signal.
        for key in self.reserved_headers + self.reserved_options:
            if kwargs is not None and key in kwargs:
                headers[key] = kwargs.pop(key)
            if key in options:
                headers[key] = options.pop(key)

        if 'headers' in options:
            if options['headers'] is None:
                options['headers'] = headers
            else:
                options['headers'].update(headers)
        else:
            options['headers'] = headers

        return super(Task, self).apply_async(
            args=args, kwargs=kwargs, task_id=task_id, producer=producer,
            link=link, link_error=link_error, shadow=shadow, serializer='girder_io', **options)

    @property
    def canceled(self):
        """
        A property to indicate if a task has been canceled.

        :returns True is this task has been canceled, False otherwise.
        """
        return is_revoked(self)

    def describe(self):
        return describe_function(self.run)

    def call_item_task(self, inputs, outputs={}):
        return self.run.call_item_task(inputs, outputs)

    def _maybe_transform_result(self, idx, result, **kwargs):
        try:
            grh = self.request.girder_result_hooks[idx]
            if hasattr(grh, 'transform') and \
               six.callable(grh.transform):
                return grh.transform(result, **kwargs)
            return result
        except IndexError:
            return result

    def _maybe_transform_argument(self, arg, **kwargs):
        if hasattr(arg, 'transform') and six.callable(arg.transform):
            return arg.transform(**kwargs)
        return arg

    def _maybe_cleanup(self, arg, **kwargs):
        if hasattr(arg, 'cleanup') and six.callable(arg.cleanup):
            arg.cleanup(**kwargs)

    def __call__(self, *args, **kwargs):
        try:
            _t_args = _walk_obj(args, self._maybe_transform_argument)
            _t_kwargs = _walk_obj(kwargs, self._maybe_transform_argument)

            results = super(Task, self).__call__(*_t_args, **_t_kwargs)

            if hasattr(self.request, 'girder_result_hooks'):
                if isinstance(results, tuple):
                    results = tuple([self._maybe_transform_result(i, r)
                                     for i, r in enumerate(results)])
                else:
                    results = self._maybe_transform_result(0, results)

            return results
        finally:
            _walk_obj(args, self._maybe_cleanup)
            _walk_obj(kwargs, self._maybe_cleanup)



import copy
import httmock
import os
import shutil
import unittest

import girder_worker
from girder_worker.utils import JobStatus

import girder_worker.core


_tmp = None


def setUpModule():
    global _tmp
    _tmp = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), 'tmp', 'fetch')
    girder_worker.config.set('girder_worker', 'tmp_root', _tmp)


def tearDownModule():
    if os.path.isdir(_tmp):
        shutil.rmtree(_tmp)


class TestIo(unittest.TestCase):
    def testDefaultInline(self):
        task = {
            'mode': 'python',
            'script': 'b = a ** 2',
            'inputs': [{
                'id': 'a',
                'format': 'number',
                'type': 'number'
            }],
            'outputs': [{
                'id': 'b',
                'format': 'number',
                'type': 'number'
            }]
        }

        # Mode should default to "inline" if data key is set
        inputs = {
            'a': {
                'data': 5,
                'format': 'number',
                'type': 'number'
            }
        }

        outputs = girder_worker.core.run(task, inputs)
        self.assertEqual(outputs['b']['data'], 25)

    def testInlineFilepath(self):
        task = {
            'mode': 'python',
            'script': """
fname = file
with open(file) as f:
    out = f.read()
""",
            'inputs': [{
                'id': 'file',
                'format': 'text',
                'type': 'string',
                'target': 'filepath'
            }],
            'outputs': [
                {
                    'id': 'out',
                    'format': 'text',
                    'type': 'string'
                },
                {
                    'id': 'fname',
                    'format': 'text',
                    'type': 'string'
                }
            ]
        }

        # Test filepath input
        inputs = {
            'file': {
                'data': 'a,b,c\n1,2,3\n',
                'format': 'text',
                'type': 'string'
            }
        }

        outputs = girder_worker.core.run(task, inputs)
        self.assertEqual(outputs['out']['data'], 'a,b,c\n1,2,3\n')

        # Test filepath input with filename specified
        task['inputs'][0]['filename'] = 'file.csv'

        inputs = {
            'file': {
                'data': 'a,b,c\n1,2,3\n',
                'format': 'text',
                'type': 'string'
            }
        }

        outputs = girder_worker.core.run(task, inputs)
        self.assertEqual(outputs['out']['data'], 'a,b,c\n1,2,3\n')
        self.assertEqual(outputs['fname']['data'][-8:], 'file.csv')

    def testHttpIo(self):
        task = {
            'mode': 'python',
            'script': 'y = x + "_suffix"\nfoo="bar"',
            'inputs': [{
                'id': 'x',
                'format': 'text',
                'type': 'string',
                'target': 'filepath',
                'filename': 'override.txt'
            }],
            'outputs': [{
                'id': 'y',
                'format': 'text',
                'type': 'string'
            }, {
                'id': 'foo',
                'format': 'text',
                'type': 'string'
            }]
        }

        inputs = {
            'x': {
                'mode': 'http',
                'url': 'https://foo.com/file.txt',
                'format': 'text',
                'type': 'string'
            }
        }

        outputs = {
            'foo': {
                'mode': 'http',
                'format': 'text',
                'type': 'string',
                'url': 'https://output.com/location.out',
                'headers': {'foo': 'bar'},
                'params': {'queryParam': 'value'},
                'method': 'PUT'
            }
        }

        job_mgr = girder_worker.utils.JobManager(
            True, url='http://jobstatus/')

        received = []
        status_changes = []

        @httmock.all_requests
        def fetchMock(url, request):
            if url.netloc == 'foo.com' and url.scheme == 'https':
                # The input fetch request
                return 'dummy file contents'
            elif url.netloc == 'output.com' and url.path == '/location.out':
                self.assertEqual(url.query, 'queryParam=value')
                received.append(request.body)
                return ''
            elif (url.netloc == 'jobstatus' and url.path == '/' and
                    request.method == 'PUT'):
                status_changes.append(request.body)
                return ''
            else:
                raise Exception('Unexpected url ' + repr(url))

        with httmock.HTTMock(fetchMock):
            # Use user-specified filename
            out = girder_worker.core.run(
                task, inputs=copy.deepcopy(inputs), outputs=outputs,
                cleanup=False, _job_manager=job_mgr, status=JobStatus.RUNNING)

            val = out['y']['data']
            self.assertTrue(val.endswith('override.txt_suffix'))
            path = val[:-7]

            # We should have received 3 status changes
            expected_statuses = [
                JobStatus.FETCHING_INPUT,
                JobStatus.RUNNING,
                JobStatus.PUSHING_OUTPUT
            ]
            self.assertEqual(status_changes, [
                'status=%d' % i for i in expected_statuses])

            self.assertTrue(os.path.exists(path), path)
            with open(path) as f:
                contents = f.read()

            self.assertEqual(contents, 'dummy file contents')

            # Make sure our output endpoint was pushed to exactly once
            self.assertEqual(len(received), 1)
            self.assertEqual(received[0], 'bar')

            # Make sure the bound output didn't get returned
            self.assertFalse('data' in out['foo'])

            # Use automatically detected filename
            del task['inputs'][0]['filename']
            out = girder_worker.core.run(
                task, inputs=copy.deepcopy(inputs), cleanup=False,
                validate=False, auto_convert=False)
            self.assertTrue(out['y']['data'].endswith('file.txt_suffix'))

            # Download to memory instead of a file (the default target value)
            del task['inputs'][0]['target']
            out = girder_worker.core.run(
                task, inputs=copy.deepcopy(inputs), cleanup=False,
                validate=False, auto_convert=False)
            self.assertEqual(out['y']['data'], 'dummy file contents_suffix')

    def testMagicVariables(self):
        task = {
            'outputs': [{
                'id': '_tempdir',
                'type': 'string',
                'format': 'text'

            }],
            'script': ''
        }

        outputs = girder_worker.core.run(task)
        self.assertTrue('_tempdir' in outputs)
        self.assertRegexpMatches(outputs['_tempdir']['data'], _tmp + '.+')

    def testConvertingStatus(self):
        job_mgr = girder_worker.utils.JobManager(
            True, url='http://jobstatus/')

        status_changes = []

        task = {
            'mode': 'python',
            'script': 'y = x',
            'inputs': [{
                'id': 'x',
                'format': 'tsv',
                'type': 'table'
            }],
            'outputs': [{
                'id': 'y',
                'format': 'tsv',
                'type': 'table'
            }]
        }

        inputs = {
            'x': {
                'mode': 'inline',
                'data': 'a,b,c\nd,e,f\n',
                'type': 'table',
                'format': 'csv'
            }
        }

        outputs = {
            'y': {
                'mode': 'inline',
                'type': 'table',
                'format': 'csv'
            }
        }

        @httmock.all_requests
        def fetchMock(url, request):
            if (url.netloc == 'jobstatus' and url.path == '/' and
                    request.method == 'PUT'):
                status_changes.append(request.body)
                return ''
            else:
                raise Exception('Unexpected url ' + repr(url))

        with httmock.HTTMock(fetchMock):
            girder_worker.core.run(
                task, inputs=inputs, outputs=outputs, _job_manager=job_mgr,
                status=JobStatus.RUNNING)

            # We should have received 3 status changes
            expected_statuses = [
                JobStatus.CONVERTING_INPUT,
                JobStatus.RUNNING,
                JobStatus.CONVERTING_OUTPUT,
                JobStatus.PUSHING_OUTPUT
            ]
            self.assertEqual(status_changes, [
                'status=%d' % i for i in expected_statuses])



def set_nested(obj, path, v):
    key = path.pop()

    if len(path) == 0:
        obj[key] = v
        return

    obj[key] = {}
    set_nested(obj[key], path, v)


output = []
for row in input['rows']:
    item = {}
    for k, v in row.iteritems():
        path = k.split('.')
        path.reverse()
        set_nested(item, path, v)
    output.append(item)



from girder_worker.core import specs
from collections import Hashable


def spec_class_generator(class_type, spec):

    """Generate a generic Task style class from a Spec style dict.
    For example:

    >>> from girder_worker.core.specs.utils import spec_class_generator
    >>> spec = {
    ...             "inputs": [
    ...                 {"name": "a",
    ...                  "type": "number",
    ...                  "format": "number"}],
    ...             "outputs": [
    ...                 {"name": "b",
    ...                  "type": "number",
    ...                  "format": "number"}],
    ...             "mode": "python",
    ...             "script": "b = a + 3"}

    # Define the class
    >>> spec_class_generator("addThree", spec)
    <class 'abc.addThree'>

    # Set a variable to hold the class
    >>> addThree = spec_class_generator("addThree", spec)

    # Instantiate the class
    >>> dict(addThree()) # doctest: +NORMALIZE_WHITESPACE
    {'inputs': [{"format": "number", "name": "a", "type": "number"}],
     'script': 'b = a + 3',
     'mode': 'python',
     'outputs': [{"format": "number", "name": "b", "type": "number"}]}
    """

    # Decorator that adds 'script' and 'mode' keywords from
    # the spec to the kw argument passed to the decorated function
    def add_spec_to_kw(func):
        def wrapped_f(self, _spec=None, **kwargs):
            for k in ['script', 'mode', 'steps', 'connections']:
                if k not in kwargs.keys():
                    try:
                        kwargs[k] = spec[k]
                    except KeyError:
                        pass
            func(self, _spec, **kwargs)
        return wrapped_f

    # __init__ function used for the class
    @add_spec_to_kw
    def __init__(self, spec, **kw):
        specs.Task.__init__(self, spec, **kw)
        for k, v in kw.items():
            self[k] = v

        return __init__

    # Define __inputs__ and __outputs__ variables on the class.
    cls_vars = {
        '__inputs__': specs.PortList(spec['inputs']),
        '__outputs__': specs.PortList(spec['outputs']),
        '__init__': __init__}

    return type(class_type, (specs.Task,), cls_vars)


def to_frozenset(item):
    """Recursively convert 'item' to a frozenset"""
    if isinstance(item, Hashable):
        return item

    if isinstance(item, dict):
        return frozenset([(k, to_frozenset(v)) for k, v in item.items()])

    if isinstance(item, list):
        return frozenset([to_frozenset(v) for v in item])



"""This module defines tasks executed in the pipeline."""

from .spec import Spec
from .port_list import PortList

from collections import MutableMapping


class ReadOnlyAttributeException(Exception):
    """Exception thrown when attempting to set a read only attribute"""
    pass


class TaskSpec(Spec):
    """Defines a pipeline element.

    A task is an element of the pipeline responsible for completing an
    atomic task given one or more inputs and pushing the result to the
    next task or tasks.
    """

    def __init__(self, *args, **kw):
        """Initialize the spec and add private attributes."""
        super(TaskSpec, self).__init__(*args, **kw)
        self._input_data = {}
        self._output_data = {}
        self._dirty = True

    def __setitem__(self, key, value):
        """Extend spec setitem to call PortList for input/output properties."""
        if key in ('inputs', 'outputs'):
            value = PortList(value)
        super(TaskSpec, self).__setitem__(key, value)

    def update(self, other, **kw):
        """Extend update to call PortList for input/output properties."""
        if 'inputs' in other:
            other['inputs'] = PortList(other['inputs'])
        if 'outputs' in other:
            other['outputs'] = PortList(other['outputs'])
        super(TaskSpec, self).update(other, **kw)

    _serializer = str


# add base spec properties
TaskSpec.make_property('mode', 'The execution mode of the task', 'python')
TaskSpec.make_property('script', 'A script or function to execute', '')
TaskSpec.make_property(
    'inputs',
    'A list of inputs accepted by the task',
    PortList()
)
TaskSpec.make_property(
    'outputs',
    'A list of outputs returned by the task',
    PortList()
)


class Task(MutableMapping):

    __inputs__ = PortList()
    __outputs__ = PortList()

    # Should probably be a property/function on Spec
    def get_spec_props(self):
        return ['mode', 'script', 'inputs', 'outputs']

    def __init__(self, spec=None, **kw):
        """Initialize the spec and add private attributes."""

        self.__spec__ = TaskSpec()

        if spec is None:
            spec = {}

        spec.update(kw)
        for k, v in spec.items():
            self[k] = v

        self.__spec__.__setitem__('inputs', self.__inputs__)
        self.__spec__.__setitem__('outputs', self.__outputs__)

    # Shadow attr getters for attributes in __spec__
    #  using Our own __getitem__ method
    def __getattr__(self, key):
        if key in self.get_spec_props():
            return self.__getitem__(key)
        else:
            return super(Task, self).__getattr__(key)

    # Shadow attr setters for attributes in __spec__
    # using our own __setitem__ method
    def __setattr__(self, key, value):
        if key in self.get_spec_props():
            return self.__setitem__(key, value)
        else:
            return super(Task, self).__setattr__(key, value)

    def __setitem__(self, key, value):
        """Extend spec setitem to call PortList for input/output properties."""

        if key in ('inputs', 'outputs'):
            raise ReadOnlyAttributeException(
                '%s is a read only attribute.' % key)
        else:
            self.__spec__.__setitem__(key, value)

    def __getitem__(self, key):
        return self.__spec__.__getitem__(key)

    def __delitem__(self, key):
        self.__spec__.__delitem__(key)

    def __iter__(self):
        return self.__spec__.__iter__()

    def __len__(self):
        return self.__spec__.__len__()

    def update(self, other=None, **kw):

        """A recursive version of ``dict.update``."""

        if other is not None:
            if {'inputs', 'outputs'} & set(other):
                raise ReadOnlyAttributeException('inputs and outputs are '
                                                 'read only attributes.')
        if {'inputs', 'outputs'} & set(kw):
            raise ReadOnlyAttributeException('inputs and outputs are '
                                             'read only attributes.')
        self.__spec__.update(other, **kw)

    def get(self, key, default=None):
        return self.__spec__.get(key, default)


__all__ = ('TaskSpec', 'Task', 'ReadOnlyAttributeException', )



from tempfile import mktemp
from girder_worker.tasks import run
import girder_worker
import os
import unittest

_tmpfiles = []


def _mockTempfile():
    global _tmpfiles
    out = mktemp()
    _tmpfiles.append(out)
    return out


girder_worker.core.executors.python.tempfile.mktemp = _mockTempfile


class TestDebug(unittest.TestCase):

    def setUp(self):
        global _tmpfiles
        _tmpfiles = []

        self.script = (
            "c = {'fields': a['fields'], 'rows': a['rows'] + b['rows']}"
        )
        self.analysis = {
            'name': 'append_tables',
            'inputs': [
                {'name': 'a', 'type': 'table', 'format': 'rows'},
                {'name': 'b', 'type': 'table', 'format': 'rows'}
            ],
            'outputs': [{'name': 'c', 'type': 'table', 'format': 'rows'}],
            'script': self.script,
            'mode': 'python'
        }

        self.inputs = {
            'a': {
                'format': 'rows.json',
                'data': ('{"fields": ["aa", "bb"], '
                         '"rows": [{"aa": 1, "bb": 2}]}')
            },
            'b': {
                'format': 'rows.json',
                'data': ('{"fields": ["aa", "bb"],'
                         '"rows": [{"aa": 3, "bb": 4}]}')
            }
        }

        self.outputs = {
            'c': {'format': 'rows.json'}
        }

        # Change dir for loading analyses
        self.prevdir = os.getcwd()
        cur_path = os.path.dirname(os.path.realpath(__file__))
        os.chdir(cur_path)
        self.analysis_path = os.path.join(cur_path, '..', 'analysis')

    def tearDown(self):
        os.chdir(self.prevdir)

    def test_analysis_debug(self):
        """Runs the table json test but with asserts for analysis debugging"""
        global _tmpfiles

        run(
            dict(self.analysis.items() + [('write_script', 1)]),
            inputs=self.inputs, outputs=self.outputs)

        # Should have generated just one debug file
        self.assertEquals(len(_tmpfiles), 1)

        # File contents equals script
        with open(_tmpfiles[0], 'r') as fh:
            self.assertEquals(fh.read(), self.script)

    def test_kwargs_debug(self):
        """Runs the table json test but with asserts for kwarg debugging"""
        global _tmpfiles

        run(
            self.analysis, self.inputs, self.outputs, write_script=True)

        # Should have generated serveral files ()
        self.assertGreater(len(_tmpfiles), 1)



from girder_worker import GirderWorkerPluginABC


class GWExamplePlugin(GirderWorkerPluginABC):
    def __init__(self, app, *args, **kwargs):
        self.app = app
        # Here we can also change application settings. E.g.
        # changing the task time limit:
        #
        # self.app.config.update({
        #     'TASK_TIME_LIMIT': 300
        # })

    def task_imports(self):
        # Return a list of python importable paths to the
        # plugin's path directory
        return ['gwexample.analyses.tasks']



from setuptools import setup

setup(name='gwexample',
      version='0.0.0',
      description='An example girder worker extension',
      author='Chris Kotfila',
      author_email='chris.kotfila@kitware.com',
      license='Apache v2',
      classifiers=[
          'Development Status :: 2 - Pre-Alpha',
          'License :: OSI Approved :: Apache Software License'
          'Topic :: Scientific/Engineering :: GIS',
          'Intended Audience :: Science/Research',
          'Natural Language :: English',
          'Programming Language :: Python'
      ],
      entry_points={
          'girder_worker_plugins': [
              'gwexample = gwexample:GWExamplePlugin',
          ]
      },
      packages=['gwexample'],
      zip_safe=False)



from __future__ import absolute_import
from . import http, local, mongodb

import os
import tempfile

_fetch_map = {}
_push_map = {}
_stream_fetch_map = {}
_stream_push_map = {}


def _inline_fetch(spec, **kwargs):
    taskInput = kwargs.get('task_input', {})
    target = taskInput.get('target', 'memory')
    if target == 'filepath':
        # Ensure we have a trailing slash
        tmpDir = os.path.join(kwargs['_tempdir'], '')

        if 'filename' in taskInput:
            filename = taskInput['filename']
            path = os.path.join(tmpDir, filename)
            with open(path, 'wb') as out:
                out.write(spec['data'])
        else:
            with tempfile.NamedTemporaryFile(
                    'wb', prefix=tmpDir, delete=False) as out:
                out.write(spec['data'])
                path = out.name

        return path
    elif target == 'memory':
        return spec['data']
    else:
        raise Exception('Invalid fetch target: ' + target)


def _inline_push(data, spec, **kwargs):
    spec['data'] = data
    return spec


def register_fetch_handler(mode, handler):
    """
    Register a new handler function for fetching data for a given mode. This
    will override any existing handler mapped to the given mode.

    :param mode: The name of the mode this handler corresponds to.
    :type mode: str
    :param handler: The handler function that performs the fetch.
    :type handler: function
    """
    _fetch_map[mode] = handler


def register_push_handler(mode, handler):
    """
    Register a new handler function for pushing data for a given mode. This
    will override any existing handler mapped to the given mode.

    :param mode: The name of the mode this handler corresponds to.
    :type mode: str
    :param handler: The handler function that performs the data push.
    :type handler: function
    """
    _push_map[mode] = handler


def register_stream_fetch_adapter(mode, adapter_cls):
    """
    Register an adapter class to handle streaming input for a given mode.

    :param mode: The mode to bind this adapter to.
    :type mode: str
    :param adapter_cls: The adapter class that will be instantiated to handle
        stream fetching for the given mode.
    :type adapter_cls: :py:class:`StreamFetchAdapter`
    """
    _stream_fetch_map[mode] = adapter_cls


def register_stream_push_adapter(mode, adapter_cls):
    """
    Register an adapter class to handle stream pushing for a given mode.

    :param mode: The mode to bind this adapter to.
    :type mode: str
    :param adapter_cls: The adapter class that will be instantiated to handle
        stream pushing for the given mode.
    :type adapter_cls: :py:class:`StreamPushAdapter`
    """
    _stream_push_map[mode] = adapter_cls


def _detect_mode(spec):
    mode = spec.get('mode', 'auto')

    if mode == 'auto':
        # We guess the mode based on the "url" value
        if 'url' not in spec:
            return 'inline'

        scheme = spec['url'].split(':', 1)[0]

        if scheme == 'https':
            mode = 'http'
        elif scheme == 'file':
            mode = 'local'
            spec['path'] = spec['url'][7:]  # Truncate "file://"
        else:
            mode = scheme

    return mode


def fetch(spec, **kwargs):
    """
    This function can be called on any valid input binding specification and is
    responsible for performing any operations necessary to fetch the data. Once
    any fetch operations are complete, this returns the value to which the
    corresponding variable should be set.

    :param input_spec: The specification of the input to fetch. This is a
        LOCATION_SPEC type in the grammar.
    :type input_spec: dict
    """
    mode = _detect_mode(spec)

    if mode not in _fetch_map:
        raise Exception('Unknown input fetch mode: ' + mode)

    return _fetch_map[mode](spec, **kwargs)


def push(data, spec, **kwargs):
    """
    The opposite of fetch, this is responsible for writing data to some
    destination in a specified mode defined by ``spec``.

    :param data: The data to push
    :type data: opaque
    :param spec: The output spec
    :type spec: dict
    """
    mode = _detect_mode(spec)

    if mode not in _push_map:
        raise Exception('Unknown output push mode: ' + mode)

    return _push_map[mode](data, spec, **kwargs)


def make_stream_fetch_adapter(input):
    """
    Create a stream fetch adapter based on the given input binding.
    """
    mode = _detect_mode(input)

    if mode not in _stream_fetch_map:
        raise Exception('Unknown streaming input fetch mode: ' + mode)

    return _stream_fetch_map[mode](input)


def make_stream_push_adapter(output):
    """
    Create a stream push adapter based on the given output binding.
    """
    mode = _detect_mode(output)

    if mode not in _stream_push_map:
        raise Exception('Unknown streaming output push mode: ' + mode)

    return _stream_push_map[mode](output)


register_fetch_handler('http', http.fetch)
register_fetch_handler('mongodb', mongodb.fetch)
register_fetch_handler('local', local.fetch)
register_fetch_handler('inline', _inline_fetch)


register_push_handler('http', http.push)
register_push_handler('mongodb', mongodb.push)
register_push_handler('local', local.push)
register_push_handler('inline', _inline_push)

register_stream_push_adapter('http', http.HttpStreamPushAdapter)
register_stream_fetch_adapter('http', http.HttpStreamFetchAdapter)



from girder_worker.tasks import run
from girder_worker.core import load
import os
import unittest


class TestWorkflow(unittest.TestCase):

    def setUp(self):
        self.add_three = {
            'inputs': [
                {
                    'name': 'a',
                    'type': 'number',
                    'format': 'number'
                }
            ],
            'outputs': [
                {
                    'name': 'b',
                    'type': 'number',
                    'format': 'number'
                }
            ],
            'mode': 'python',
            'script': 'b = a + 3'
        }

        self.add_two = {
            'inputs': [
                {
                    'name': 'a',
                    'type': 'number',
                    'format': 'number'
                }
            ],
            'outputs': [
                {
                    'name': 'b',
                    'type': 'number',
                    'format': 'number'
                }
            ],
            'mode': 'python',
            'script': 'b = a + 2'
        }

        self.add = {
            'inputs': [
                {
                    'name': 'a',
                    'type': 'number',
                    'format': 'number',
                },
                {
                    'name': 'b',
                    'type': 'number',
                    'format': 'number'
                }
            ],
            'outputs': [
                {
                    'name': 'c',
                    'type': 'number',
                    'format': 'number'
                }
            ],
            'script': 'c = a + b',
            'mode': 'python'
        }

        self.multiply = {
            'inputs': [
                {
                    'name': 'in1',
                    'type': 'number',
                    'format': 'number'
                },
                {
                    'name': 'in2',
                    'type': 'number',
                    'format': 'number'
                }
            ],
            'outputs': [
                {
                    'name': 'out',
                    'type': 'number',
                    'format': 'number'
                }
            ],
            'mode': 'python',
            'script': 'out = in1 * in2'
        }

        self.workflow = {
            'mode': 'workflow',
            'inputs': [
                {
                    'name': 'x',
                    'type': 'number',
                    'format': 'number',
                    'default': {'format': 'number', 'data': 10}
                },
                {
                    'name': 'y',
                    'type': 'number',
                    'format': 'number'
                }
            ],
            'outputs': [
                {
                    'name': 'result',
                    'type': 'number',
                    'format': 'number'
                }
            ],
            'steps': [
                {
                    'name': 'af352b243109c4235d2549',
                    'task': self.add_three,
                },
                {
                    'name': 'af352b243109c4235d25fb',
                    'task': self.add_two,
                },
                {
                    'name': 'af352b243109c4235d25ec',
                    'task': self.multiply,
                }
            ],
            'connections': [
                {
                    'name': 'x',
                    'input_step': 'af352b243109c4235d2549',
                    'input': 'a'
                },
                {
                    'name': 'y',
                    'input_step': 'af352b243109c4235d25fb',
                    'input': 'a'
                },
                {
                    'output_step': 'af352b243109c4235d2549',
                    'output': 'b',
                    'input_step': 'af352b243109c4235d25ec',
                    'input': 'in1'
                },
                {
                    'output_step': 'af352b243109c4235d25fb',
                    'output': 'b',
                    'input_step': 'af352b243109c4235d25ec',
                    'input': 'in2'
                },
                {
                    'name': 'result',
                    'output_step': 'af352b243109c4235d25ec',
                    'output': 'out'
                }
            ]
        }

        self.multi_input = {
            'mode': 'workflow',
            'inputs': [
                {
                    'name': 'x',
                    'type': 'number',
                    'format': 'number'
                },
                {
                    'name': 'y',
                    'type': 'number',
                    'format': 'number'
                }
            ],
            'outputs': [
                {
                    'name': 'result',
                    'type': 'number',
                    'format': 'number'
                }
            ],
            'steps': [
                {
                    'name': 1,
                    'task': self.add,
                },
                {
                    'name': 2,
                    'task': self.multiply,
                },
                {
                    'name': 3,
                    'task': self.multiply,
                }
            ],
            'connections': [
                {
                    'name': 'x',
                    'input_step': 2,
                    'input': 'in1'
                },
                {
                    'name': 'x',
                    'input_step': 2,
                    'input': 'in2'
                },
                {
                    'name': 'y',
                    'input_step': 3,
                    'input': 'in1'
                },
                {
                    'name': 'y',
                    'input_step': 3,
                    'input': 'in2'
                },
                {
                    'output_step': 2,
                    'output': 'out',
                    'input_step': 1,
                    'input': 'a'
                },
                {
                    'output_step': 3,
                    'output': 'out',
                    'input_step': 1,
                    'input': 'b'
                },
                {
                    'name': 'result',
                    'output_step': 1,
                    'output': 'c'
                }
            ]
        }

        self.visualization = {
            'mode': 'workflow',
            'inputs': [
                {
                    'name': 'x',
                    'type': 'number',
                    'format': 'number'
                },
                {
                    'name': 'y',
                    'type': 'number',
                    'format': 'number'
                }
            ],
            'outputs': [
                {
                    'name': 'result',
                    'type': 'number',
                    'format': 'number'
                }
            ],
            'steps': [
                {
                    'name': 1,
                    'task': self.add,
                },
                {
                    'name': 2,
                    'task': self.multiply,
                },
                {
                    'name': 3,
                    'task': self.multiply,
                },
                {
                    'name': 4,
                    'visualization': True,
                    'task': {
                        'inputs': [
                            {
                                'name': 'x',
                                'type': 'number',
                                'format': 'number'
                            }
                        ]
                    }
                }
            ],
            'connections': [
                {
                    'name': 'x',
                    'input_step': 2,
                    'input': 'in1'
                },
                {
                    'name': 'x',
                    'input_step': 2,
                    'input': 'in2'
                },
                {
                    'name': 'y',
                    'input_step': 3,
                    'input': 'in1'
                },
                {
                    'name': 'y',
                    'input_step': 3,
                    'input': 'in2'
                },
                {
                    'output_step': 2,
                    'output': 'out',
                    'input_step': 1,
                    'input': 'a'
                },
                {
                    'output_step': 3,
                    'output': 'out',
                    'input_step': 1,
                    'input': 'b'
                },
                {
                    'name': 'result',
                    'output_step': 1,
                    'output': 'c'
                },
                {
                    'output_step': 1,
                    'output': 'c',
                    'input_step': 4,
                    'input': 'x'
                }
            ]
        }
        # Change dir for loading analyses
        self.prevdir = os.getcwd()
        cur_path = os.path.dirname(os.path.realpath(__file__))
        os.chdir(cur_path)
        self.analysis_path = os.path.join(cur_path, '..', 'analysis')

    def tearDown(self):
        os.chdir(self.prevdir)

    def test_workflow(self):
        outputs = run(
            self.workflow,
            inputs={
                'x': {'format': 'json', 'data': '1'},
                'y': {'format': 'number', 'data': 2}
            })
        self.assertEqual(outputs['result']['format'], 'number')
        self.assertEqual(outputs['result']['data'], (1+3)*(2+2))

        # Test using the default value for x (10).
        outputs = run(
            self.workflow,
            inputs={
                'y': {'format': 'number', 'data': 2}
            })
        self.assertEqual(outputs['result']['format'], 'number')
        self.assertEqual(outputs['result']['data'], (10+3)*(2+2))

    def test_multi_input(self):
        outputs = run(
            self.multi_input,
            inputs={
                'x': {'format': 'number', 'data': 2},
                'y': {'format': 'number', 'data': 3}
            })
        self.assertEqual(outputs['result']['format'], 'number')
        self.assertEqual(outputs['result']['data'], (2*2)+(3*3))

    def test_visualization(self):
        outputs = run(
            self.visualization,
            inputs={
                'x': {'format': 'number', 'data': 2},
                'y': {'format': 'number', 'data': 3}
            })
        self.assertEqual(outputs['result']['format'], 'number')
        self.assertEqual(outputs['result']['data'], (2*2)+(3*3))
        self.assertEqual(outputs['_visualizations'], [{
            'mode': 'preset',
            'type': 4,
            'inputs': {
                'x': {
                    'format': 'number',
                    'type': 'number',
                    'data': (2*2)+(3*3)
                }
            }
        }])

    def test_load(self):
        flu = load(os.path.join(
            self.analysis_path, 'xdata', 'flu.json'))

        script_path = os.path.join(self.analysis_path, 'xdata', 'flu.py')
        with open(script_path) as script_file:
            script_content = script_file.read()

        self.assertEqual(flu['script'], script_content)

    def test_load_workflow(self):
        flu = load(os.path.join(
            self.analysis_path, 'xdata', 'flu_workflow.json'))

        script_path = os.path.join(self.analysis_path, 'xdata', 'flu.py')
        with open(script_path) as script_file:
            script_content = script_file.read()

        self.assertEqual(flu['steps'][0]['task']['script'], script_content)


if __name__ == '__main__':
    unittest.main()



from ete3 import Tree


def populate_tree(node, children):
    for c in children:
        name = c.get('node_data', {}).get('node name', '')
        dist = c.get('edge_data', {}).get('weight')
        n = node.add_child(name=name, dist=dist)
        populate_tree(n, c.get('children', []))


t = Tree()
populate_tree(t, input.get('children', []))
output = t.write(format=1)



def fetch(spec, **kwargs):
    """
    Fetches a file on the local filesystem into memory.
    """
    with open(spec['path'], 'rb') as f:
        return f.read()


def push(data, spec, **kwargs):
    """
    Write a blob of data in memory to a file specified in ``spec['path']``.
    """
    with open(spec['path'], 'wb') as out:
        out.write(data)



"""This module defines the workflow pipeline object."""

from girder_worker.core.specs import PortList, ReadOnlyAttributeException, Spec

from collections import MutableMapping
import networkx as nx
import copy


class WorkflowException(Exception):
    """Exception thrown for issues with Workflows"""
    pass


class DuplicateTaskException(WorkflowException):
    """Exception thrown when adding a duplicate task"""
    pass


class StepSpec(Spec):
    def __init__(self, *args, **kwargs):
        super(StepSpec, self).__init__(*args, **kwargs)


StepSpec.make_property('name', 'Name of the step in the workflow')
StepSpec.make_property('task', 'Workflow task associated with this step')


class ConnectionSpec(Spec):
    def __init__(self, *args, **kwargs):
        super(ConnectionSpec, self).__init__(*args, **kwargs)


ConnectionSpec.make_property(
    'name', 'Workflow external facing input or output name')
ConnectionSpec.make_property('input', 'Name of input Port')
ConnectionSpec.make_property('input_step', 'Name of input step')
ConnectionSpec.make_property('output', 'Name of input Port')
ConnectionSpec.make_property('output_step', 'Name of output step')


class Workflow(MutableMapping):

    def __init__(self, *args, **kw):
        self.__graph__ = nx.DiGraph()

        self.__interface = {'mode', 'steps', 'connections',
                            'inputs', 'outputs'}
        self._defaults = {}

        self.mode = 'workflow'

    ##
    # Graph API
    #

    def _add_node(self, name, task):
        if name in self.__graph__.nodes():
            raise DuplicateTaskException('%s is already in Workflow!' % name)
        self.__graph__.add_node(name, task=task)

    def add_task(self, task, name=None):
        if hasattr(task, 'name'):
            self._add_node(task.name, task)
        elif name is not None:
            self._add_node(name, task)
        else:
            raise TypeError('If task does not have a name attribute, '
                            'name argument cannot be None.')
        pass

    def add_tasks(self, tasks):
        for task in tasks:
            if isinstance(task, tuple):
                self.add_task(*task)
            else:
                self.add_task(task)

    def _add_edge(self, t1, t2, metadata):
        # metadata must be a dict
        # TODO: add validation for arg0 and arg1 inputs/outputs
        self.__graph__.add_edge(t1, t2, metadata)

    def connect_tasks(self, *args, **kwargs):
        """
        Connect two tasks together specifying which input ports connect
        to which output ports. This function takes either 1, 2 or 3 arguments.
        In the case of one argument it expects a list of tuples where the tuples
        are the first, second and third arguments to connect_tasks. If there are
        two arguments given they must be nodes in the workflow and ports must
        be specified as key word arguments.  If three arguments are supplied
        they should be node1, node2 and a dict of port connections where keys
        are the output ports on node1 and value are the input ports on node2.
        """

        # Single iterable argument - apply connect_tasks() to
        # each element in the iterable
        if len(args) == 1 and hasattr(args[0], '__iter__'):
            for arg in args[0]:
                self.connect_tasks(*arg)

        # Two arguments,  assume metadata is contained in kwargs
        # _add_edge will raise an exception if this is not the case
        elif len(args) == 2:
            kwargs = kwargs if kwargs is not None else {}
            self._add_edge(args[0], args[1], kwargs)

        # Three arguments,  assume these are t1, t2 and a dict of edge data
        elif len(args) == 3:
            # Update metadata dict with key word arguments if kwargs is not None
            if kwargs is not None:
                args[2].update(kwargs)

            self._add_edge(*args)
        else:
            raise TypeError(
                'connect_tasks() takes either 3 arguments or an '
                'iterable of 3 argument tuples. (%s given)' % len(args))
    ##
    # Dict-like API
    #

    @property
    def steps(self):
        return [StepSpec({'name': n,  'task': dict(data['task'])})
                for n, data in self.__graph__.nodes(data=True)]

    def _get_connections_and_open_ports(self):
        """Internal function that returns three arguments. The first argument
        is a list of tuples of the form (output_node, output_port, input_node,
        input_port).  The second argument is a dict of open input ports. The
        third argument is a dict of open output ports.
        """

        # Generate sets of input/output ports indexed by node name
        # We will remove ports from these input/output sets as we
        # loop through the graph edge data.  This should leave us
        # with all 'open' input ports and all 'open' output ports
        input_ports = {}
        output_ports = {}
        for name, data in self.__graph__.nodes(data=True):
            input_ports[name] = set([d['name'] for d in data['task']['inputs']])
            output_ports[name] = set(
                [d['name'] for d in data['task']['outputs']])

        internal_connections = []
        for output_node, input_node, connections in self.__graph__.edges(
                data=True):
            for output_port, input_port in connections.items():

                # Add to internal connections set
                internal_connections.append((output_node, output_port,
                                             input_node, input_port))

                # Pop output_port off of output_node
                output_ports[output_node].remove(output_port)

                # Pop input_port off of input_node
                input_ports[input_node].remove(input_port)

        return internal_connections, input_ports, output_ports

    def _get_node_name(self, node, port, ports):
        """
        Return 'node.port' if there is any other port under a different node
        with the same name. Otherwise return 'port'.
        """
        if any(port in S for k, S in ports.items() if k != node):
            return '%s.%s' % (node, port)
        return port

    @property
    def connections(self):
        internal, input_ports, output_ports = (
            self._get_connections_and_open_ports())
        connections = []

        for node, inputs in input_ports.items():
            for input_port in inputs:
                connections.append(
                    ConnectionSpec({
                        'name': self._get_node_name(
                            node, input_port, input_ports),
                        'input_step': node,
                        'input': input_port
                    }))

        for output_node, output_port, input_node, input_port in internal:
            connections.append(
                ConnectionSpec({
                    'output_step': output_node,
                    'output': output_port,
                    'input_step': input_node,
                    'input': input_port
                }))

        for node, outputs in output_ports.items():
            for output_port in outputs:
                connections.append(
                    ConnectionSpec({
                        'name': self._get_node_name(
                            node, output_port, output_ports),
                        'output_step': node,
                        'output': output_port
                    }))

        return connections

    @property
    def inputs(self):
        _, open_port_names, _ = self._get_connections_and_open_ports()
        ports = []

        for node, port_names in open_port_names.items():

            # Hash the actual analysis imports by name
            input_ports = {v['name']: v for v in
                           self.__graph__.node[node]['task']['inputs']}

            # loop through open port names for this node
            for port_name in port_names:

                # Copy the actual analysis port
                port = copy.deepcopy(input_ports[port_name])

                # Set the port name (handles conflicts by prefixing with
                # node name)
                port['name'] = self._get_node_name(
                    node, port_name, open_port_names)

                # Check defaults
                if self.has_default(port['name']):
                    port['default'] = self.get_default(port['name'])

                ports.append(port)

        return PortList(ports)

    @property
    def outputs(self):
        _, _, open_port_names = self._get_connections_and_open_ports()
        ports = []

        for node, port_names in open_port_names.items():

            # Hash the actual analysis imports by name
            output_ports = {v['name']: v for v in
                            self.__graph__.node[node]['task']['outputs']}

            # loop through open port names for this node
            for port_name in port_names:

                # Copy the actual analysis port
                port = copy.deepcopy(output_ports[port_name])

                # Set the port name (handles conflicts by prefixing with
                # node name)
                port['name'] = self._get_node_name(
                    node, port_name, open_port_names)

                # Check defaults
                if self.has_default(port['name']):
                    port['default'] = self.get_default(port['name'])

                ports.append(port)

        return PortList(ports)

    def __getitem__(self, key):
        if key not in self.__interface:
            raise KeyError('Workflow keys must be one of %s' %
                           ', '.join(self.__interface))

        return getattr(self, key)

    def __setitem__(self, key, value):
        if key in self.__interface:
            raise ReadOnlyAttributeException(
                '%s should not be directly assigned')

        raise KeyError('Workflow keys must be one of %s' %
                       ', '.join(self.__interface))

    def __delitem__(self, key):
        if key in self.__interface:
            raise ReadOnlyAttributeException(
                '%s should not be directly removed')
        raise KeyError(key)

    def __len__(self):
        return len(self.__interface)

    def __iter__(self):
        for key in self.__interface:
            yield key

    #####
    #  Workflow API
    #

    # Implementing a full default wrapper because we may need
    # to support more sophisticated functionality down the line.
    # For example we may want to 'alais' external workflow port names
    # to internal open port names. This means we could set a default
    # on 'a2.a' and then later alais 'a2.a'  to 'x' the input port
    # would still need to look like {name: 'x', default: {....}, ...}
    # even though we set the default on 'a2.a.' to support something like
    # that we need a consistent CRUD-like API around self._defaults.

    def set_default(self, node_name, default):
        # TODO create DefaultSpec and wrap assignment here
        self._defaults[node_name] = default

    def get_default(self, node_name):
        return self._defaults[node_name]

    def has_default(self, node_name):
        return node_name in self._defaults.keys()

    def remove_default(self, node_name):
        try:
            del self._defaults[node_name]
        except KeyError:
            pass


__all__ = ('Workflow', 'StepSpec', 'ConnectionSpec')



from .spec import Spec
from .port import Port, ValidationError
from .port_list import PortList
from .task import Task, TaskSpec, ReadOnlyAttributeException
from .workflow import Workflow, StepSpec, ConnectionSpec
from .workflow import WorkflowException, DuplicateTaskException

__all__ = ('Spec', 'Port', 'PortList', 'ValidationError',
           'TaskSpec', 'Task', 'ReadOnlyAttributeException',
           'Workflow', 'StepSpec', 'ConnectionSpec',
           'WorkflowException', 'DuplicateTaskException')



from networkx import NetworkXNoPath

from girder_worker import utils
from girder_worker.core import events, io, run, set_job_status
from .format import conv_graph, converter_path, get_validator_analysis, Validator


def isvalid(type, binding, fetch=True, **kwargs):
    """
    Determine whether a data binding is of the appropriate type and format.

    :param type: The expected type specifier string of the binding.
    :param binding: A binding dict of the form
        ``{'format': format, 'data', data}``, where ``format`` is the format
        specifier string, and ``data`` is the raw data to test.
        The dict may also be of the form
        ``{'format': format, 'uri', uri}``, where ``uri`` is the location of
        the data (see :py:mod:`girder_worker.uri` for URI formats).
    :param fetch: Whether to do an initial data fetch before conversion
        (default ``True``).
    :returns: ``True`` if the binding matches the type and format,
        ``False`` otherwise.
    """
    kwargs = kwargs.copy()
    kwargs['auto_convert'] = False
    kwargs['validate'] = False

    analysis = get_validator_analysis(Validator(type, binding['format']))
    outputs = run(analysis, {'input': binding}, fetch=fetch, **kwargs)
    return outputs['output']['data']


def convert(type, input, output, fetch=True, status=None, **kwargs):
    """
    Convert data from one format to another.

    :param type: The type specifier string of the input data.
    :param input: A binding dict of the form
        ``{'format': format, 'data', data}``, where ``format`` is the format
        specifier string, and ``data`` is the raw data to convert.
        The dict may also be of the form
        ``{'format': format, 'uri', uri}``, where ``uri`` is the location of
        the data (see :py:mod:`girder_worker.uri` for URI formats).
    :param output: A binding of the form
        ``{'format': format}``, where ``format`` is the format
        specifier string to convert the data to.
        The binding may also be in the form
        ``{'format': format, 'uri', uri}``, where ``uri`` specifies
        where to place the converted data.
    :param fetch: Whether to do an initial data fetch before conversion
        (default ``True``).
    :returns: The output binding
        dict with an additional field ``'data'`` containing the converted data.
        If ``'uri'`` is present in the output binding, instead saves the data
        to the specified URI and
        returns the output binding unchanged.
    """
    kwargs = kwargs.copy()
    kwargs['auto_convert'] = False

    if fetch:
        input['data'] = io.fetch(input, **kwargs)

    if input['format'] == output['format']:
        data = input['data']
    else:
        data_descriptor = input
        try:
            conversion_path = converter_path(
                Validator(type, input['format']), Validator(type, output['format']))
        except NetworkXNoPath:
            raise Exception('No conversion path from %s/%s to %s/%s' % (
                type, input['format'], type, output['format']))

        # Run data_descriptor through each conversion in the path
        for conversion in conversion_path:
            result = run(
                conversion, {'input': data_descriptor}, status=status, **kwargs)
            data_descriptor = result['output']
        data = data_descriptor['data']

    if status == utils.JobStatus.CONVERTING_OUTPUT:
        job_mgr = kwargs.get('_job_manager')
        set_job_status(job_mgr, utils.JobStatus.PUSHING_OUTPUT)
    io.push(data, output, **kwargs)
    return output


def handle_input(e):
    task_input = e.info['task_input']
    input = e.info['input']
    name = e.info['name']
    kwargs = e.info['info']['kwargs']
    auto_convert = kwargs.get('auto_convert', True)
    validate = kwargs.get('validate', True)

    # Validate the input
    if validate and not isvalid(
            task_input['type'], input,
            **dict({'task_input': task_input, 'fetch': False}, **kwargs)):
        raise Exception(
            'Input %s (Python type %s) is not in the expected type (%s) and format (%s).' % (
                name, type(input['data']), task_input['type'], input['format']))

    # Convert data
    if auto_convert:
        try:
            converted = convert(
                task_input['type'], input, {'format': task_input['format']},
                status=utils.JobStatus.CONVERTING_INPUT,
                **dict({'task_input': task_input, 'fetch': False}, **kwargs))
        except Exception, e:
            raise Exception('%s: %s' % (name, str(e)))

        input['script_data'] = converted['data']
    elif not validate or input.get('format', task_input.get('format')) == task_input.get('format'):
        input['script_data'] = input['data']
    else:
        raise Exception(
            'Expected exact format match but \'%s != %s\'.' % (
                input['format'], task_input['format']))


def handle_output(e):
    task_output = e.info['task_output']
    outputs = e.info['outputs']
    name = e.info['name']
    kwargs = e.info['info']['kwargs']
    output = outputs[name]
    auto_convert = kwargs.get('auto_convert', True)
    validate = kwargs.get('validate', True)

    script_output = {
        'data': output['script_data'],
        'format': task_output.get('format')
    }
    if 'format' not in output:
        output['format'] = script_output['format']

    if validate and not isvalid(
            task_output['type'], script_output,
            **dict({'task_output': task_output}, **kwargs)):
        raise Exception(
            'Output %s (%s) is not in the expected type (%s) and format (%s).' % (
                name, type(script_output['data']), task_output['type'], output['format']))

    if auto_convert:
        outputs[name] = convert(
            task_output['type'], script_output, output, status=utils.JobStatus.CONVERTING_OUTPUT,
            **dict({'task_output': task_output}, **kwargs))
        e.prevent_default()  # convert handles the push for us
    elif validate and outputs[name]['format'] != task_output['format']:
        raise Exception(
            'Expected format match but %s != %s.' % (output['format'], task_output['format']))


def load(params):
    from girder_worker.app import app

    @app.task(name='girder_worker.convert')
    def _convert(*pargs, **kwargs):
        return convert(*pargs, **kwargs)

    @app.task(name='girder_worker.validators')
    def _validators(*pargs, **kwargs):
        type, format = pargs
        nodes = []

        for (node, data) in conv_graph.nodes(data=True):
            if type in (None, node.type) and format in (None, node.format):
                nodes.append({
                    'type': node.type,
                    'format': node.format,
                    'validator': data
                })

        return nodes

    events.bind('run.handle_input', params['name'], handle_input)
    events.bind('run.handle_output', params['name'], handle_output)



from girder_worker.utils import _maybe_model_repr
from girder_worker_utils import _walk_obj


def create_task_job(job_defaults, sender=None, body=None,
                    exchange=None, routing_key=None, headers=None,
                    properties=None, declare=None, retry_policy=None,
                    **kwargs):

    from girder.utility.model_importer import ModelImporter
    from girder.plugins.worker import utils
    from girder.api.rest import getCurrentUser

    job_model = ModelImporter.model('job', 'jobs')

    user = headers.pop('girder_user', getCurrentUser())

    # Sanitize any Transform objects
    task_args = tuple(_walk_obj(body[0], _maybe_model_repr))
    task_kwargs = _walk_obj(body[1], _maybe_model_repr)

    job = job_model.createJob(
        **{'title': headers.pop('girder_job_title',
                                job_defaults.get('girder_job_title', '')),
           'type': headers.pop('girder_job_type',
                               job_defaults.get('girder_job_type', '')),
           'handler': headers.pop('girder_job_handler',
                                  job_defaults.get('girder_job_handler', '')),
           'public': headers.pop('girder_job_public',
                                 job_defaults.get('girder_job_public', '')),
           'user': user,
           'args': task_args,
           'kwargs': task_kwargs,
           'otherFields': dict(
               celeryTaskId=headers['id'],
               **headers.pop('girder_job_other_fields',
                             job_defaults.get('girder_job_other_fields', '')))})

    headers['jobInfoSpec'] = utils.jobInfoSpec(job)
    return job


def attach_girder_api_url(sender=None, body=None, exchange=None,
                          routing_key=None, headers=None, properties=None,
                          declare=None, retry_policy=None, **kwargs):
    from girder.plugins.worker import utils
    headers['girder_api_url'] = utils.getWorkerApiUrl()


def attach_girder_client_token(sender=None, body=None, exchange=None,
                               routing_key=None, headers=None, properties=None,
                               declare=None, retry_policy=None, **kwargs):
    from girder.utility.model_importer import ModelImporter
    from girder.api.rest import getCurrentUser
    token_model = ModelImporter.model('token')
    scope = 'jobs.rest.create_job'
    try:
        token = token_model.createToken(scope=scope, user=user)
    except NameError:
        token = token_model.createToken(scope=scope, user=getCurrentUser())
    headers['girder_client_token'] = token['_id']


def get_async_result_job_property(async_result):
    # GirderAsyncResult() objects may be instantiated in
    # either a girder REST request, or in some other
    # context (e.g. from a running girder_worker instance
    # if there is a chain).  If we are in a REST request
    # we should have access to the girder package and can
    # directly access the database If we are in a
    # girder_worker context (or even in python console or
    # a testing context) then we should get an ImportError
    # and we can make a REST request to get the
    # information we need.
    from girder.utility.model_importer import ModelImporter
    job_model = ModelImporter.model('job', 'jobs')

    try:
        return job_model.findOne({'celeryTaskId': async_result.task_id})
    except IndexError:
        return None



import contextlib
import errno
import functools
import imp
import json
import os
import girder_worker
import girder_worker.plugins
import select
import shutil
import six
import subprocess
import stat
import sys
import tempfile
import traceback


class TerminalColor(object):
    """
    Provides a set of values that can be used to color text in the terminal.
    """
    ERROR = '\033[1;91m'
    SUCCESS = '\033[32m'
    WARNING = '\033[1;33m'
    INFO = '\033[35m'
    ENDC = '\033[0m'

    @staticmethod
    def _color(tag, text):
        return ''.join((tag, text, TerminalColor.ENDC))

    @staticmethod
    def error(text):
        return TerminalColor._color(TerminalColor.ERROR, text)

    @staticmethod
    def success(text):
        return TerminalColor._color(TerminalColor.SUCCESS, text)

    @staticmethod
    def warning(text):
        return TerminalColor._color(TerminalColor.WARNING, text)

    @staticmethod
    def info(text):
        return TerminalColor._color(TerminalColor.INFO, text)


def toposort(data):
    """
    General-purpose topological sort function. Dependencies are expressed as a
    dictionary whose keys are items and whose values are a set of dependent
    items. Output is a list of sets in topological order. This is a generator
    function that returns a sequence of sets in topological order.

    :param data: The dependency information.
    :type data: dict
    :returns: Yields a list of sorted sets representing the sorted order.
    """
    if not data:
        return

    # Ignore self dependencies.
    for k, v in data.items():
        v.discard(k)

    # Find all items that don't depend on anything.
    extra = functools.reduce(
        set.union, data.itervalues()) - set(data.iterkeys())
    # Add empty dependences where needed
    data.update({item: set() for item in extra})

    # Perform the toposort.
    while True:
        ordered = set(item for item, dep in data.iteritems() if not dep)
        if not ordered:
            break
        yield ordered
        data = {item: (dep - ordered)
                for item, dep in data.iteritems() if item not in ordered}
    # Detect any cycles in the dependency graph.
    if data:
        raise Exception('Cyclic dependencies detected:\n%s' % '\n'.join(
                        repr(x) for x in data.iteritems()))


@contextlib.contextmanager
def tmpdir(cleanup=True):
    # Make the temp dir underneath tmp_root config setting
    root = os.path.abspath(girder_worker.config.get(
        'girder_worker', 'tmp_root'))
    try:
        os.makedirs(root)
    except OSError:
        if not os.path.isdir(root):
            raise
    path = tempfile.mkdtemp(dir=root)

    try:
        yield path
    finally:
        # Cleanup the temp dir
        if cleanup and os.path.isdir(path):
            shutil.rmtree(path)


def with_tmpdir(fn):
    """
    This function is provided as a convenience to allow use as a decorator of
    a function rather than using "with tmpdir()" around the whole function
    body. It passes the generated temp dir path into the function as the
    special kwarg "_tempdir".
    """
    @functools.wraps(fn)
    def wrapped(*args, **kwargs):
        if '_tempdir' in kwargs:
            return fn(*args, **kwargs)

        cleanup = kwargs.get('cleanup', True)
        with tmpdir(cleanup=cleanup) as tempdir:
            kwargs['_tempdir'] = tempdir
            return fn(*args, **kwargs)
    return wrapped


class PluginNotFoundException(Exception):
    pass


def load_plugins(plugins, paths, ignore_errors=False, quiet=False):
    """
    Enable a list of plugins.

    :param plugins: The plugins to enable.
    :type plugins: list or tuple of str
    :param paths: Plugin search paths.
    :type paths: list or tuple of str
    :param ignore_errors: If a plugin fails to load, this determines whether to
        raise the exception or simply print an error and keep going.
    :type ignore_errors: bool
    :param quiet: Optionally suppress printing status messages.
    :type quiet: bool
    :return: Set of plugins that were loaded successfully.
    :rtype: set
    """
    loaded = set()
    for plugin in plugins:
        try:
            load_plugin(plugin, paths)
            loaded.add(plugin)
            if not quiet:
                print(TerminalColor.success('Loaded plugin "%s"' % plugin))
        except Exception:
            print(TerminalColor.error(
                'ERROR: Failed to load plugin "%s":' % plugin))
            if ignore_errors:
                traceback.print_exc()
            else:
                raise

    return loaded


def load_plugin(name, paths):
    """
    Enable a plugin for the worker runtime.

    :param name: The name of the plugin to load, which is also the name of its
        containing directory.
    :type name: str
    :param paths: Plugin search paths.
    :type paths: list or tuple of str
    """
    for path in paths:
        plugin_dir = os.path.join(path, name)
        if os.path.isdir(plugin_dir):
            module_name = 'girder_worker.plugins.' + name

            if module_name not in sys.modules:
                fp, pathname, description = imp.find_module(name, [path])
                module = imp.load_module(module_name, fp, pathname, description)
                setattr(girder_worker.plugins, name, module)
            else:
                module = sys.modules[module_name]

            if hasattr(module, 'load'):
                module.load({
                    'plugin_dir': plugin_dir,
                    'name': name
                })

            break
    else:
        raise PluginNotFoundException(
            'Plugin "%s" not found. Looked in: \n   %s\n' % (
                name, '\n   '.join(paths)))


def _close_pipes(rds, wds, input_pipes, output_pipes, close_output_pipe):
    """
    Helper to close remaining input and output adapters after the subprocess
    completes.
    """
    # close any remaining output adapters
    for fd in rds:
        if fd in output_pipes:
            output_pipes[fd].close()
            if close_output_pipe(fd):
                os.close(fd)

    # close any remaining input adapters
    for fd in wds:
        if fd in input_pipes:
            os.close(fd)


def _setup_input_pipes(input_pipes):
    """
    Given a mapping of input pipes, return a tuple with 2 elements. The first is
    a list of file descriptors to pass to ``select`` as writeable descriptors.
    The second is a dictionary mapping paths to existing named pipes to their
    adapters.
    """
    wds = []
    fifos = {}
    for pipe, adapter in six.viewitems(input_pipes):
        if isinstance(pipe, int):
            # This is assumed to be an open system-level file descriptor
            wds.append(pipe)
        else:
            if not os.path.exists(pipe):
                raise Exception('Input pipe does not exist: %s' % pipe)
            if not stat.S_ISFIFO(os.stat(pipe).st_mode):
                raise Exception('Input pipe must be a fifo object: %s' % pipe)
            fifos[pipe] = adapter

    return wds, fifos


def _open_ipipes(wds, fifos, input_pipes):
    """
    This will attempt to open the named pipes in the set of ``fifos`` for
    writing, which will only succeed if the subprocess has opened them for
    reading already. This modifies and returns the list of write descriptors,
    the list of waiting fifo names, and the mapping back to input adapters.
    """
    for fifo in fifos.copy():
        try:
            fd = os.open(fifo, os.O_WRONLY | os.O_NONBLOCK)
            input_pipes[fd] = fifos.pop(fifo)
            wds.append(fd)
        except OSError as e:
            if e.errno != errno.ENXIO:
                raise e

    return wds, fifos, input_pipes


def select_loop(exit_condition=lambda: True, close_output=lambda x: True,
                outputs=None, inputs=None):
    """
    Run a select loop for a set of input and output pipes

    :param exit_condition: A function to evaluate to determine if the select
        loop should terminate if all pipes are empty.
    :type exit_condition: function
    :param close_output: A function to use to test whether a output
        should be closed when EOF is reached. Certain output pipes such as
        stdout, stderr should not be closed.
    :param outputs: This should be a dictionary mapping pipe descriptors
        to instances of ``StreamPushAdapter`` that should handle the data from
        the stream. The keys of this dictionary are open file descriptors,
        which are integers.
    :type outputs: dict
    :param inputs: This should be a dictionary mapping pipe descriptors
        to instances of ``StreamFetchAdapter`` that should handle sending
        input data in chunks. Keys in this dictionary can be either open file
        descriptors (integers) or a string representing a path to an existing
        fifo on the filesystem. This second case supports the use of named
        pipes, since they must be opened for reading before they can be opened
        for writing
    :type inputs: dict
    """

    BUF_LEN = 65536
    inputs = inputs or {}
    outputs = outputs or {}

    rds = [fd for fd in outputs.keys() if isinstance(fd, int)]
    wds, fifos = _setup_input_pipes(inputs)

    try:
        while True:
            # We evaluate this first so that we get one last iteration of
            # of the loop before breaking out of the loop.
            exit = exit_condition()

            # get ready pipes, timeout of 100 ms
            readable, writable, _ = select.select(rds, wds, (), 0.1)

            for ready_fd in readable:
                buf = os.read(ready_fd, BUF_LEN)

                if buf:
                    outputs[ready_fd].write(buf)
                else:
                    outputs[ready_fd].close()
                    # Should we close this pipe? In the case of stdout or stderr
                    # bad things happen if parent closes
                    if close_output(ready_fd):
                        os.close(ready_fd)
                    rds.remove(ready_fd)
            for ready_fd in writable:
                # TODO for now it's OK for the input reads to block since
                # input generally happens first, but we should consider how to
                # support non-blocking stream inputs in the future.
                buf = inputs[ready_fd].read(BUF_LEN)

                if buf:
                    os.write(ready_fd, buf)
                else:   # end of stream
                    wds.remove(ready_fd)
                    os.close(ready_fd)

            wds, fifos, inputs = _open_ipipes(wds, fifos, inputs)
            # all pipes empty?
            empty = (not rds or not readable) and (not wds or not writable)

            if (empty and exit):
                break

    finally:
        _close_pipes(rds, wds, inputs, outputs, close_output)


def run_process(command, output_pipes=None, input_pipes=None):
    """
    Run a subprocess, and listen for its outputs on various pipes.

    :param command: The command to run.
    :type command: list of str
    :param output_pipes: This should be a dictionary mapping pipe descriptors
        to instances of ``StreamPushAdapter`` that should handle the data from
        the stream. Normally, keys of this dictionary are open file descriptors,
        which are integers. There are two special cases where they are not,
        which are the keys ``'_stdout'`` and ``'_stderr'``. These special keys
        correspond to the stdout and stderr pipes that will be created for the
        subprocess. If these are not set in the ``output_pipes`` map, the
        default behavior is to direct them to the stdout and stderr of the
        current process.
    :type output_pipes: dict
    :param input_pipes: This should be a dictionary mapping pipe descriptors
        to instances of ``StreamFetchAdapter`` that should handle sending
        input data in chunks. Keys in this dictionary can be either open file
        descriptors (integers), the special value ``'_stdin'`` for standard
        input, or a string representing a path to an existing fifo on the
        filesystem. This third case supports the use of named pipes, since they
        must be opened for reading before they can be opened for writing
    :type input_pipes: dict
    """

    p = subprocess.Popen(args=command, stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE, stdin=subprocess.PIPE)

    input_pipes = input_pipes or {}
    output_pipes = output_pipes or {}

    # we now know subprocess stdout and stderr filenos, so bind the adapters
    stdout = p.stdout.fileno()
    stderr = p.stderr.fileno()
    stdin = p.stdin.fileno()
    output_pipes[stdout] = output_pipes.get(
        '_stdout', WritePipeAdapter({}, sys.stdout))
    output_pipes[stderr] = output_pipes.get(
        '_stderr', WritePipeAdapter({}, sys.stderr))

    # Special case for _stdin
    if '_stdin' in input_pipes:
        input_pipes[stdin] = input_pipes['_stdin']

    def exit_condition():
        status = p.poll()
        return status is not None

    def close_output_pipe(pipe):
        return pipe not in (stdout, stderr)

    try:
        select_loop(exit_condition=exit_condition,
                    close_output=close_output_pipe,
                    outputs=output_pipes, inputs=input_pipes)
    except Exception:
        p.kill()  # kill child process if something went wrong on our end
        raise

    return p


class StreamFetchAdapter(object):
    """
    This represents the interface that must be implemented by fetch adapters
    for IO modes that want to implement streaming input.
    """
    def __init__(self, input_spec):
        self.input_spec = input_spec

    def read(self, buf_len):
        """
        Fetch adapters must implement this method, which is responsible for
        reading up to ``self.buf_len`` bytes from the stream. For now, this is
        expected to be a blocking read, and should return an empty string to
        indicate the end of the stream.
        """
        raise NotImplemented


class MemoryFetchAdapter(StreamFetchAdapter):
    def __init__(self, input_spec, data):
        """
        Simply reads data from memory. This can be used to map traditional
        (non-streaming) inputs to pipes when using ``run_process``. This is
        roughly identical behavior to BytesIO.
        """
        super(MemoryFetchAdapter, self).__init__(input_spec)
        self._stream = six.BytesIO(data)

    def read(self, buf_len):
        return self._stream.read(buf_len)


class StreamPushAdapter(object):
    """
    This represents the interface that must be implemented by push adapters for
    IO modes that want to implement streaming output.
    """
    def __init__(self, output_spec):
        """
        Initialize the adpater based on the output spec.
        """
        self.output_spec = output_spec

    def write(self, buf):
        """
        Write a chunk of data to the output stream.
        """
        raise NotImplemented

    def close(self):
        """
        Close the output stream. Called after the last data is sent.
        """
        pass


class WritePipeAdapter(StreamPushAdapter):
    """
    Simply wraps another pipe that contains a ``write`` method. This is useful
    for wrapping ``sys.stdout`` and ``sys.stderr``, where we want to call
    ``write`` but not ``close`` on them.
    """
    def __init__(self, output_spec, pipe):
        """
        :param pipe: An object containing a ``write`` method, e.g. sys.stdout.
        """
        super(WritePipeAdapter, self).__init__(output_spec)
        self.pipe = pipe

    def write(self, buf):
        self.pipe.write(buf)


class AccumulateDictAdapter(StreamPushAdapter):
    def __init__(self, output_spec, key, dictionary=None):
        """
        Appends all data from a stream under a key inside a dict. Can be used
        to bind traditional (non-streaming) outputs to pipes when using
        ``run_process``.

        :param output_spec: The output specification.
        :type output_spec: dict
        :param key: The key to accumulate the data under.
        :type key: hashable
        :param dictionary: Dictionary to write into. If not specified, uses the
            output_spec.
        :type dictionary: dict
        """
        super(AccumulateDictAdapter, self).__init__(output_spec)

        if dictionary is None:
            dictionary = output_spec

        if key not in dictionary:
            dictionary[key] = ''

        self.dictionary = dictionary
        self.key = key

    def write(self, buf):
        self.dictionary[self.key] += buf


class JobProgressAdapter(StreamPushAdapter):
    def __init__(self, job_manager):
        """
        This reads structured JSON documents one line at a time and sends
        them as progress events via the JobManager.

        :param job_manager: The job manager to use to send the progress events.
        :type job_manager: girder_worker.utils.JobManager
        """
        super(JobProgressAdapter, self).__init__(None)

        self.job_manager = job_manager
        self._buf = b''

    def write(self, buf):
        lines = buf.split(b'\n')
        if self._buf:
            lines[0] = self._buf + lines[0]
        self._buf = lines[-1]

        for line in lines[:-1]:
            self._parse(line)

    def _parse(self, line):
        try:
            doc = json.loads(line.decode('utf8'))
        except ValueError:
            return  # TODO log?

        if not isinstance(doc, dict):
            return  # TODO log?

        self.job_manager.updateProgress(
            total=doc.get('total'), current=doc.get('current'), message=doc.get('message'))



import os


def load(params):
    from girder_worker.core import register_executor
    from girder_worker.plugins.types import format
    from . import executor

    register_executor('r', executor.run)

    converters_dir = os.path.join(params['plugin_dir'], 'converters')
    format.import_converters([
        os.path.join(converters_dir, 'r'),
        os.path.join(converters_dir, 'table'),
        os.path.join(converters_dir, 'tree')
    ])



from girder_worker.tasks import run
import unittest


class TestIntegerList(unittest.TestCase):

    def setUp(self):
        self.analysis = {
            'name': 'concatenate',
            'inputs': [
                {
                    'name': 'a',
                    'type': 'integer_list',
                    'format': 'integer_list',
                    'default': {
                        'format': 'json',
                        'data': '[0]'
                    }
                },
                {
                    'name': 'b',
                    'type': 'integer_list',
                    'format': 'integer_list'
                }
            ],
            'outputs': [{'name': 'c',
                         'type': 'integer_list', 'format': 'integer_list'}],
            'script': 'c = a + b',
            'mode': 'python'
        }

    def test_integer(self):
        outputs = run(
            self.analysis,
            inputs={
                'a': {'format': 'integer_list', 'data': [1, 2]},
                'b': {'format': 'integer_list', 'data': [3, 4]}
            },
            outputs={
                'c': {'format': 'integer_list'}
            })
        self.assertEqual(outputs['c']['format'], 'integer_list')
        self.assertEqual(outputs['c']['data'], [1, 2, 3, 4])

    def test_json(self):

        outputs = run(
            self.analysis,
            inputs={
                'a': {'format': 'json', 'data': '[1, 2]'},
                'b': {'format': 'json', 'data': '[3, 4]'}
            },
            outputs={
                'c': {'format': 'json'}
            })
        self.assertEqual(outputs['c']['format'], 'json')
        self.assertEqual(outputs['c']['data'], '[1, 2, 3, 4]')

    def test_default(self):
        outputs = run(
            self.analysis,
            inputs={
                'b': {'format': 'integer_list', 'data': [2]}
            },
            outputs={
                'c': {'format': 'integer_list'}
            })
        self.assertEqual(outputs['c']['format'], 'integer_list')
        self.assertEqual(outputs['c']['data'], [0, 2])

        self.assertRaisesRegexp(
            Exception, '^Required input \'b\' not provided.$',
            run, self.analysis,
            inputs={
                'a': {'format': 'integer_list', 'data': [2]}
            },
            outputs={
                'c': {'format': 'integer_list'}
            })


if __name__ == '__main__':
    unittest.main()



import errno
import os
import shutil
from girder_worker_utils.transform import Transform
from girder_worker_utils.transforms.girder_io import (
    GirderUploadJobArtifact,
    GirderClientTransform,
    GirderUploadToFolder,
    GirderUploadToItem
)
from . import TemporaryVolume, Connect, NamedOutputPipe, _maybe_transform


class ProgressPipe(Transform):

    def __init__(self, name='.girder_progress', volume=TemporaryVolume.default):
        self.name = name
        self._volume = volume

    def transform(self, task=None, **kwargs):
        from girder_worker.docker.stream_adapter import JobProgressAdapter

        self._volume.transform(**kwargs)

        # TODO What do we do is job_manager is None? When can it me None?
        job_mgr = task.job_manager
        # For now we are using JobProgressAdapter which is part of core, we should
        # probably add a copy to the docker package to break to reference to core.
        return Connect(NamedOutputPipe(self.name, volume=self._volume),
                       JobProgressAdapter(job_mgr)).transform(**kwargs)


class GirderFileIdToStream(GirderClientTransform):
    def __init__(self, _id, **kwargs):
        super(GirderFileIdToStream, self).__init__(**kwargs)
        self.file_id = _id

    def transform(self, **kwargs):
        from girder_worker.docker.io.girder import (
            GirderFileStreamReader
        )
        return GirderFileStreamReader(self.gc, self.file_id)


class GirderFileIdToVolume(GirderClientTransform):
    def __init__(self, _id, volume=TemporaryVolume.default, filename=None, **kwargs):
        super(GirderFileIdToVolume, self).__init__(**kwargs)
        self._file_id = str(_id)
        self._volume = volume
        self._filename = filename
        self._file_path = None

    def _create_file_path(self, root):
        if self._filename is None:
            # If no filename is explicitly passed, we read the filename from Girder
            # and put it in its own directory named by its UUID.
            filename = self.gc.getFile(self._file_id)['name']
            path = os.path.join(root, self._file_id)
            try:
                os.mkdir(path)
            except OSError as e:
                if e.errno != errno.EEXIST:
                    raise

            return os.path.join(self._file_id, filename), os.path.join(path, filename)
        else:
            return self._filename, os.path.join(root, self._filename)

    def transform(self, **kwargs):
        self._volume.transform(**kwargs)
        dir = self._volume.host_path
        rel_path, self._file_path = self._create_file_path(dir)

        self.gc.downloadFile(self._file_id, self._file_path)

        # Return the path inside the container
        return os.path.join(self._volume.container_path, rel_path)

    def cleanup(self, **kwargs):
        if self._file_path is not None:
            shutil.rmtree(self._file_path, ignore_errors=True)

    def _repr_model_(self):
        if self._filename:
            template = u'<{module}.{cls}: File ID={id} -> "{fname}">'
        else:
            template = u'<{module}.{cls}: File ID={id}>'
        return template.format(
            module=self.__module__, cls=self.__class__.__name__, id=self._file_id,
            fname=self._filename)


class GirderFolderIdToVolume(GirderClientTransform):
    def __init__(self, _id, volume=TemporaryVolume.default, folder_name=None, **kwargs):
        super(GirderFolderIdToVolume, self).__init__(**kwargs)
        self._folder_id = str(_id)
        self._volume = volume
        self._folder_name = folder_name
        self._folder_path = None

    def _create_folder_path(self, root):
        if self._folder_name is None:
            # If no folder name is explicitly passed, we read the filename from Girder
            # and put it in its own directory named by its UUID.
            self._folder_name = self.gc.getFolder(self._folder_id)['name']
        path = os.path.join(root, self._folder_id, self._folder_name)
        try:
            os.makedirs(path)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise

        return os.path.join(self._folder_id, self._folder_name), path

    def transform(self, **kwargs):
        self._volume.transform(**kwargs)
        dir = self._volume.host_path
        rel_path, self._folder_path = self._create_folder_path(dir)

        self.gc.downloadFolderRecursive(self._folder_id, self._folder_path)

        # Return the path inside the container
        return os.path.join(self._volume.container_path, rel_path)

    def cleanup(self, **kwargs):
        if self._folder_path is not None:
            shutil.rmtree(self._folder_path, ignore_errors=True)

    def _repr_model_(self):
        return '<%s.%s: Folder ID=%s -> "%s">' % (
            self.__module__, self.__class__.__name__, self._folder_id, self._folder_path)


class GirderUploadVolumePathToItem(GirderUploadToItem):
    def __init__(self, volumepath, item_id, delete_file=False, **kwargs):
        item_id = str(item_id)
        super(GirderUploadVolumePathToItem, self).__init__(item_id, delete_file, **kwargs)
        self._volumepath = volumepath

    # We ignore the "result"
    def transform(self, *args, **kwargs):
        path = _maybe_transform(self._volumepath, *args, **kwargs)

        return super(GirderUploadVolumePathToItem, self).transform(path)


class GirderUploadVolumePathToFolder(GirderUploadToFolder):
    def __init__(self, volumepath, folder_id, delete_file=False, **kwargs):
        super(GirderUploadVolumePathToFolder, self).__init__(str(folder_id), delete_file, **kwargs)
        self._volumepath = volumepath

    def transform(self, *args, **kwargs):
        path = _maybe_transform(self._volumepath, *args, **kwargs)
        return super(GirderUploadVolumePathToFolder, self).transform(path)


class GirderUploadVolumePathJobArtifact(GirderUploadJobArtifact):
    def __init__(self, volumepath, job_id=None, name=None, **kwargs):
        if job_id is not None:
            job_id = str(job_id)
        super(GirderUploadVolumePathJobArtifact, self).__init__(job_id, name, **kwargs)
        self._volumepath = volumepath

    def transform(self, *args, **kwargs):
        path = _maybe_transform(self._volumepath, *args, **kwargs)
        return super(GirderUploadVolumePathJobArtifact, self).transform(path)
